# SPDX-FileCopyrightText: 2024-2025 ModelCloud.ai
# SPDX-FileCopyrightText: 2024-2025 qubitium@modelcloud.ai
# SPDX-License-Identifier: Apache-2.0
# Contact: qubitium@modelcloud.ai, x.com/qubitium

from .base_qwen3_vl import BaseQwen3VLGPTQ


class Qwen3_VLQModel(BaseQwen3VLGPTQ):
    pass
