# SPDX-FileCopyrightText: 2024-2025 ModelCloud.ai
# SPDX-FileCopyrightText: 2024-2025 qubitium@modelcloud.ai
# SPDX-License-Identifier: Apache-2.0
# Contact: qubitium@modelcloud.ai, x.com/qubitium

# odd minor versions are dev (main) branch
# even minor versions are release
# 5.2.0 => release, 5.1.0 => devel
# micro version (5.2.x) denotes patch fix, i.e. 5.2.1 is a patch fix release
__version__ = "5.8.0"
